from coalib.misc.Decorators import generate_ordering, generate_repr


@generate_repr("match", "range")
@generate_ordering("range", "match")
class Match:
    """
    Stores information about a single textual match.
    """

    def __init__(self, match, position):
        """
        Instantiates a new Match.

        :param match:    The actual matched string.
        :param position: The position where the match was found. Starts from
                         zero.
        """
        self._match = match
        self._position = position

    def __len__(self):
        return len(self.match)

    def __str__(self):
        return self.match

    @property
    def match(self):
        """
        Returns the text matched.

        :returns: The text matched.
        """
        return self._match

    @property
    def position(self):
        """
        Returns the position where the text was matched (zero-based).

        :returns: The position.
        """
        return self._position

    @property
    def end_position(self):
        """
        Marks the end position of the matched text (zero-based).

        :returns: The end-position.
        """
        return len(self) + self.position

    @property
    def range(self):
        """
        Returns the position range where the text was matched.

        :returns: A pair indicating the position range. The first element is
                  the start position, the second one the end position.
        """
        return (self.position, self.end_position)



class CircularDependencyError(Exception):

    @classmethod
    def for_bears(cls, bears):
        """
        Creates the CircularDependencyError with a helpful message about the
        dependency.
        """
        bear_names = [bear.name for bear in bears]

        return cls("Circular dependency detected: " + " -> ".join(bear_names))


def _resolve(bears, resolved_bears, seen):
    for bear in bears:
        if bear in resolved_bears:
            continue

        missing = bear.missing_dependencies(resolved_bears)
        if missing == []:
            resolved_bears.append(bear)
            continue

        if bear in seen:
            seen.append(bear)
            raise CircularDependencyError.for_bears(seen)

        seen.append(bear)
        resolved_bears = _resolve(missing, resolved_bears, seen)
        resolved_bears.append(bear)
        seen.remove(bear)  # Already resolved, no candidate for circular dep

    return resolved_bears


def resolve(bears):
    """
    Collects all dependencies of the given bears. This will also remove
    duplicates.

    :param bears: The given bears. Will not be modified.
    :return:      The new list of bears, sorted so that it can be executed
                  sequentially without dependency issues.
    """
    return _resolve(bears, [], [])
